"""
Graph Interface

A set of functions to graph data. Matplotlib and plotly are supported.
"""

#We use just-in-time importing here to improve load times
#Here are the imports:
#import plotly.express as px
#import matplotlib.pyplot as plt
from MLVisualizationTools.types import GraphDataTypes, GraphOutputTypes, ColorizerModes
from MLVisualizationTools import Colorizers 
from MLVisualizationTools.backend import GraphData
from typing import Optional
import copy
import pandas as pd


class WrongDataFormatException(Exception):
    pass

def graph(data: GraphData, graphtype: GraphOutputTypes = GraphOutputTypes.Auto, title="", legend: bool =True,
          sizekey: Optional[str] = None):
    """
    Calls correct graph type based on data format and mode chosen
    
    :param data: GraphData from interface call
    :param graphtype: Matplotlib or plotly
    :param title: Title for graph
    :param legend: Show a key for the colors used
    :param sizekey: Should match with datainterface sizekey if used
    """

    if graphtype == GraphOutputTypes.Auto:
        try:
            import plotly
            return plotlyGraph(data, title, legend, sizekey)
        except ImportError:
            try:
                import matplotlib
                return matplotlibGraph(data, title, legend, sizekey)
            except ImportError:
                raise ImportError("Either matplotlib or plotly is required to use graphs. Install them with pip.")

    elif graphtype == GraphOutputTypes.Plotly:
        return plotlyGraph(data, title, legend, sizekey)
    elif graphtype == GraphOutputTypes.Matplotlib:
        return matplotlibGraph(data, title, legend, sizekey)
    else:
        raise Exception(f"GraphType {graphtype} not recognized.")

def plotlyGraph(data: GraphData, title="", legend: bool=True, sizekey: Optional[str] = None):
    """
    Calls correct graph type based on data format
    
    :param data: GraphData from interface call
    :param title: Title for graph
    :param legend: Show a key for the colors used
    :param sizekey: Should match with datainterface sizekey if used
    """
    if data.datatype is GraphDataTypes.Grid:
        return plotlyGrid(data, title, legend, sizekey)
    elif data.datatype is GraphDataTypes.Animation:
        return plotlyAnimation(data, title, legend, sizekey)
    else:
        raise Exception(f"DataType {data.datatype} not recognized.")

def matplotlibGraph(data: GraphData, title="", legend: bool=True, sizekey: Optional[str] = None):
    """
    Calls correct graph type based on data format
    
    :param data: GraphData from interface call
    :param title: Title for graph
    :param legend: Show a key for the colors used
    :param sizekey: Should match with datainterface sizekey if used
    """
    
    if data.datatype is GraphDataTypes.Grid:
        return matplotlibGrid(data, title, legend, sizekey)
    elif data.datatype is GraphDataTypes.Animation:
        return matplotlibAnimation(data, title, legend, sizekey) #Unsupported
    else:
        raise Exception(f"DataType {data.datatype} not recognized.")


def _applyDefaultFormat(data: GraphData):
    """Internal function used by graphs to make datapoints stand out"""
    if data.modeldata.colorized == ColorizerModes.NotColorized:
        data = Colorizers.simple(data, 'blue', apply_to_model=True, apply_to_data=False)
    if data.datavalues is not None:
        if data.datavalues.colorized == ColorizerModes.NotColorized:
            data = Colorizers.simple(data, 'black', apply_to_data=True, apply_to_model=False)

        df = data.datavalues.dataframe
        df[data.sizekey] = df[data.sizekey].apply(lambda x: x * 50 / df[data.sizekey].max())

    data.modeldata.dataframe[data.sizekey] = 5
    return data

def _plotlyGraphCore(data: GraphData, title, legend):
    """Internal function for plotly graphing, users should call plotlyGraph() instead"""
    try:
        import plotly.express as px
    except ImportError:
        raise ImportError("Plotly is required to use this graph. Install with `pip install plotly`")
    
    sm = 50 if data.datavalues is not None else None

    cdm = {}
    order = []
    locations = [data.modeldata]
    if data.datavalues is not None:
        locations.append(data.datavalues)

    for datavals in locations:
        for colordata in [datavals.truecolor, datavals.falsecolor, datavals.basecolor]:
            if colordata.color is not None:
                cdm[colordata.message] = colordata.color
                order.append(colordata.message)
                
                df = datavals.dataframe
                df.loc[df[data.colorkey] == colordata.color, data.colorkey] = colordata.message

    if data.datavalues is not None:
        combined_df = pd.concat([data.modeldata.dataframe, data.datavalues.dataframe])
    else:
        combined_df = data.modeldata.dataframe

    fig = px.scatter_3d(combined_df, data.x, data.y, data.outputkey, animation_frame=data.anim, color=data.colorkey,
                        color_discrete_map=cdm, category_orders={data.colorkey: order}, opacity=1, size=data.sizekey,
                        title=title, range_z=[combined_df[data.outputkey].min(), combined_df[data.outputkey].max()],
                        size_max=sm)

    fig.update_layout(showlegend=legend and len(order) > 0)
    fig.update_traces(marker={'line_width': 0})
    return fig

def plotlyGrid(data: GraphData, title="", legend: bool=True, sizekey: Optional[str] = None):
    """
    Calls px.scatter_3d with data. Returns a plotly figure.

    :param data: GraphData from interface call
    :param title: Title for graph
    :param legend: Show a key for the colors used
    :param sizekey: Should match with datainterface sizekey if used
    """
    if data.datatype != GraphDataTypes.Grid:
        raise WrongDataFormatException("Data was not formatted in grid.")
    
    if sizekey is not None:
        data.sizekey = sizekey
    data.check_size_key()

    data = _applyDefaultFormat(copy.deepcopy(data))

    # Consistency with plotly animation key style
    locations = [data.modeldata]
    if data.datavalues is not None:
        locations.append(data.datavalues)
    for datavals in locations:
        orig_row = datavals.dataframe.iloc[0]
        for color in [datavals.basecolor, datavals.truecolor, datavals.falsecolor]:  # apply each color
            if color.color is not None:
                row = copy.deepcopy(orig_row)
                row[data.colorkey] = color.color
                row[data.sizekey] = 0
                datavals.dataframe = pd.concat([datavals.dataframe, pd.DataFrame.from_records([row])])

    return _plotlyGraphCore(data, title, legend)

def plotlyAnimation(data: GraphData, title="", legend: bool=True, sizekey: Optional[str] = None):
    """
    Calls px.scatter_3d with data and animation frame. Returns a plotly figure.

    :param data: GraphData from interface call
    :param title: Title for graph
    :param legend: Show a key for the colors used
    :param sizekey: Key to use in dataframe to store size values, should match with datainterface sizekey if used
    """
    if data.datatype != GraphDataTypes.Animation:
        raise WrongDataFormatException("Data was not formatted in animation.")
    
    if sizekey is not None:
        data.sizekey = sizekey
    data.check_size_key()

    data = _applyDefaultFormat(copy.deepcopy(data))

    # plotly animations have a bug where points aren't rendered unless
    # one point of each color is in frame
    locations = [data.modeldata]
    if data.datavalues is not None:
        locations.append(data.datavalues)
    for datavals in locations:
        orig_row = datavals.dataframe.iloc[0]
        for animval in data.modeldata.dataframe[data.anim].unique(): #apply to each frame (use model data for smoother animation)
            for color in [datavals.basecolor, datavals.truecolor, datavals.falsecolor]: #apply each color
                if color.color is not None:
                    row = copy.deepcopy(orig_row)
                    row[data.colorkey] = color.color
                    row[data.sizekey] = 0
                    row[data.anim] = animval
                    datavals.dataframe = pd.concat([datavals.dataframe, pd.DataFrame.from_records([row])])

    return _plotlyGraphCore(data, title, legend)

def matplotlibGrid(data: GraphData, title="", legend: bool = True, sizekey: Optional[str] = None):
    """
    Calls ax.scatter with data. Returns a plt instance, a fig, and the ax.

    :param data: GraphData from interface call
    :param title: Title for graph
    :param legend: Should show legend
    :param sizekey: Unused in this graph form
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
    except ImportError:
        raise ImportError("Matplotlib is required to use this graph. Install with `pip install matplotlib`")

    if data.datatype != GraphDataTypes.Grid:
        raise WrongDataFormatException("Data was not formatted in grid.")

    if sizekey is not None:
        data.sizekey = sizekey
    data.check_size_key()

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    data = _applyDefaultFormat(copy.deepcopy(data))

    if legend:
        patches = []
        locations = [data.modeldata]
        if data.datavalues is not None:
            locations.append(data.datavalues)

        for datavals in locations:
            if datavals.colorized == ColorizerModes.Binary:
                patches.append(mpatches.Patch(color=datavals.truecolor.color, label=datavals.truecolor.message))
                patches.append(mpatches.Patch(color=datavals.falsecolor.color, label=datavals.falsecolor.message))

            if datavals.colorized == ColorizerModes.Simple:
                patches.append(mpatches.Patch(color=datavals.basecolor.color, label=datavals.basecolor.message))

        if len(patches) > 1:
            ax.legend(handles=patches)

    df = data.modeldata.dataframe
    ax.scatter(df[data.x], df[data.y], df[data.outputkey], c=df[data.colorkey])

    if data.datavalues is not None:
        df = data.datavalues.dataframe
        ax.scatter(df[data.x], df[data.y], df[data.outputkey], s=df[data.sizekey], c=df[data.colorkey])

    ax.set_xlabel(data.x)
    ax.set_ylabel(data.y)
    ax.set_zlabel(data.outputkey)
    ax.set_title(title)

    return plt, fig, ax

def matplotlibAnimation(_data: GraphData, _title="", _legend: bool = True, _sizekey: Optional[str] = None):
    """This function is not implemented and is not planned to be implemented in the future."""
    raise NotImplementedError("Matplotib does not support animations cleanly. This is not planned to be added in the "
                              "future. Use plotly instead (`pip install plotly`)")


"""
Model Interfaces

A set of functions to generate graphable data on how a ML model responds to input conditions.

Analytics functions can be called directly with a pandas dataframe, or indirectly through
the raw alternative, which takes a colinfo object.

Grid (3D) and Animation (4D) options are provided.
"""

from MLVisualizationTools.backend import colinfo, GraphData, GraphDataTypes
from MLVisualizationTools.modelanalytics import AnalyticsColumnInfo
from typing import List, Dict, Union
import pandas as pd
import warnings

#Functions for passing data to ml models

#region grid
def predictionGrid(model, x:Union[str, AnalyticsColumnInfo], y:Union[str, AnalyticsColumnInfo],
                   data:pd.DataFrame, exclude:List[str] = None, steps:int=20,
                   outputkey: str = 'Output') -> GraphData:
    """
    Creates a dataset from a 2d prediction on a ML model. Wrapper function for PredictionGridRaw()
    that automatically handles column info generation.

    :param model: A ML model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    :param outputkey: Used to override default output name
    """
    coldata, allcols = colinfo(data, exclude)
    return predictionGridRaw(model, x, y, coldata, allcols, steps, outputkey)

def predictionGridRaw(model, x:Union[str, AnalyticsColumnInfo], y:Union[str, AnalyticsColumnInfo],
                      coldata:Dict[str, Dict], allcols: List[str], steps:int=20,
                      outputkey: str = 'Output') -> GraphData:
    """
    Creates a dataset from a 2d prediction on a ML model.

    Call from Grid to autogen params.

    Coldata should be formatted with keys 'min', 'max', 'mean'

    :param model: A ML model
    :param model: A ML model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param coldata: An dict of dicts accessed by col names with min, max, and mean values
    :param allcols: Ordered list of column names
    :param steps: Resolution to scan model with
    :param outputkey: Used to override default output name
    """
    if isinstance(x, AnalyticsColumnInfo):
        x = x.name
    if isinstance(y, AnalyticsColumnInfo):
        y = y.name

    assert x in coldata, "X must be in coldata"
    assert y in coldata, "Y must be in coldata"

    cols = []
    srow = []
    for name, item in coldata.items():
        if name not in [x, y]:
            cols.append(name)
        srow.append(item['mean'])

    srow = [srow] * (steps ** 2)
    preddata = pd.DataFrame(srow, columns=allcols)

    col = []
    for pos in range(0, steps):
        col.append(round(pos * (coldata[x]['max'] - coldata[x]['min']) / (steps - 1) + coldata[x]['min'], 6))
    col = col * steps
    preddata[x] = col

    col = []
    for pos in range(0, steps):
        col += [round(pos * (coldata[y]['max'] - coldata[y]['min']) / (steps - 1) + coldata[y]['min'], 6)] * steps
    preddata[y] = col

    predictions = model.predict(preddata)
    if outputkey in preddata.columns:
        warnings.warn(f"Output key '{outputkey}' was already in dataframe. This means that '{outputkey}' "
                      "was a key in your dataset and could result in data being overwritten. "
                      "You can pick a different key in the function call.")
    preddata[outputkey] = predictions
    return GraphData(preddata, GraphDataTypes.Grid, steps, x, y, outputkey=outputkey)
#endregion grid

#region animation
def predictionAnimation(model, x:Union[str, AnalyticsColumnInfo], y:Union[str, AnalyticsColumnInfo],
                        anim:Union[str, AnalyticsColumnInfo], data: pd.DataFrame, exclude:List[str] = None,
                        steps:int=20, outputkey: str = 'Output') -> GraphData:
    """
    Creates a dataset from a 2d prediction on a ML model. Wrapper function for PredictionGridRaw()
    that automatically handles column info generation.

    :param model: A ML model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param anim: Animation axis for graph data
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    :param outputkey: Used to override default output name
    """
    coldata, allcols = colinfo(data, exclude)
    return predictionAnimationRaw(model, x, y, anim, coldata, allcols, steps, outputkey)

def predictionAnimationRaw(model, x:Union[str, AnalyticsColumnInfo], y:Union[str, AnalyticsColumnInfo],
                           anim:Union[str, AnalyticsColumnInfo], coldata:Dict[str, Dict], allcols: List[str],
                           steps:int=20, outputkey: str = 'Output') -> GraphData:
    """
    Creates a dataset from a 2d prediction on a ML model.

    Call from PredictionAnimation to autogen params.

    Coldata should be formatted with keys 'name', 'min', 'max', 'mean'

    :param model: A ML model
    :param model: A ML model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param anim: Animation axis for graph data
    :param coldata: An dict of dicts accessed by col names with min, max, and mean values
    :param allcols: Ordered list of column names
    :param steps: Resolution to scan model with
    :param outputkey: Used to override default output name
    """

    if isinstance(x, AnalyticsColumnInfo):
        x = x.name
    if isinstance(y, AnalyticsColumnInfo):
        y = y.name
    if isinstance(anim, AnalyticsColumnInfo):
        anim = anim.name

    assert x in coldata, "X must be in coldata"
    assert y in coldata, "Y must be in coldata"
    assert anim in coldata, "Anim must be in coldata"

    cols = []
    srow = []
    for name, item in coldata.items():
        if name not in [x, y, anim]:
            cols.append(name)
        srow.append(item['mean'])

    srow = [srow] * (steps ** 3)
    preddata = pd.DataFrame(srow, columns=allcols)

    col = []
    for pos in range(0, steps):
         col.append(round(pos * (coldata[x]['max'] - coldata[x]['min']) / (steps - 1) + coldata[x]['min'], 6))
    col = col * (steps ** 2)
    preddata[x] = col

    col = []
    for pos in range(0, steps):
        col += [round(pos * (coldata[y]['max'] - coldata[y]['min']) / (steps - 1) + coldata[y]['min'], 6)] * steps
    col = col * steps
    preddata[y] = col

    col = []
    for pos in range(0, steps):
        col += [round(pos * (coldata[anim]['max'] - coldata[anim]['min']) / (steps - 1) + coldata[anim]['min'],6 )] * (steps ** 2)
    preddata[anim] = col

    predictions = model.predict(preddata)
    if outputkey in preddata.columns:
        warnings.warn(f"Output key '{outputkey}' was already in dataframe. This means that '{outputkey}' "
                      "was a key in your dataset and could result in data being overwritten. "
                      "You can pick a different key in the function call.")
    preddata[outputkey] = predictions
    return GraphData(preddata, GraphDataTypes.Animation, steps, x, y, anim, outputkey)

#endregion



from MLVisualizationTools.backend import fileloader
import pandas as pd

# This code was used to prep the training data for the demos based on the titanic dataset
# The orig file was left in train_orig.csv
# The modified file was moved to train.csv
# All operations performed are standard preprocessing

def main():
    df = pd.read_csv(fileloader('examples/Datasets/Titanic/train_orig.csv'))
    df.drop(['PassengerId', 'Name', 'Ticket', 'Cabin'], axis=1, inplace=True) #these cols don't have much useful data
    df['Age'].fillna(df['Age'].mean(), inplace=True) #this data has lots of nan values
    df = df[df['Embarked'].notna()] #this col is only missing two values
    df["Embarked"].replace({"C": 0, "S": 1, "Q": 2}, inplace=True) #numerical values are better for training
    df['Sex'].replace({"male": 0, "female": 1}, inplace=True)
    df.to_csv(fileloader('examples/Datasets/Titanic/train.csv'), index=False)

if __name__ == '__main__':
    main()


import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' #stops agressive error message printing
from tensorflow import keras
import pandas as pd
from MLVisualizationTools.backend import fileloader

# This creates a keras model for use in demo visualizations
# It uses the Datasets/Titanic/train.csv file
# Limited preprocessing is applied
# Model is trained with 50 epochs and batch size of 10
# Evaluation is applied at the end

def getModel():
    model = keras.Sequential()
    model.add(keras.layers.Dense(10, input_dim=7, activation='relu'))
    model.add(keras.layers.Dense(10, activation='relu'))
    model.add(keras.layers.Dense(1, activation='sigmoid'))
    model.compile(loss="binary_crossentropy", optimizer="adam", metrics=["accuracy"])
    return model

def main():
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))

    #region preprocess
    header = list(df.columns)
    header.remove("Survived")

    X = df[header].values
    Y = df["Survived"].values

    splitpoint = int(len(df)*0.7)
    trainX = X[:splitpoint]
    testX = X[splitpoint:]

    trainY = Y[:splitpoint]
    testY = Y[splitpoint:]
    #endregion

    print("Creating Model")
    model = getModel()
    print("Starting Training")
    model.fit(trainX, trainY, epochs=50, batch_size=10)
    print("Starting Evaluation")
    _, trainaccuracy = model.evaluate(trainX,trainY)
    _, testaccuracy = model.evaluate(testX, testY)
    print()
    print("We achieved an (training) accuracy of:", str(round(trainaccuracy,3) * 100) + "%")
    print("We achieved an (testing) accuracy of:", str(round(testaccuracy,3) * 100) + "%")
    print()
    model.save(fileloader('examples/Models/titanicmodel'))
    print("Model saved to Models/titanicmodel")

if __name__ == '__main__':
    main()


from werkzeug.serving import is_running_from_reloader

def main(theme = 'dark', highcontrast = True):
    """
    Runs the demo by calling DashModelVisualizer

    :param theme: theme, could be 'light' or 'dark'
    :param highcontrast: Use blue and orange coloring instead of red and green
    """
    from MLVisualizationTools.express import DashModelVisualizer
    from MLVisualizationTools.backend import fileloader
    import pandas as pd
    import os
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # stops agressive error message printing
    from tensorflow import keras
    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))
    df = df.drop("Survived", axis=1)
    DashModelVisualizer.visualize(model, df, title="DashInteractiveDemo", theme=theme, highcontrast=highcontrast)

if not is_running_from_reloader() and __name__ != '__mp_main__':
    print("This demo is for use outside of a jupyter notebook and uses the default precompiled model.")
    print("To run the demo, call DashDemo.main()")

if __name__ == "__main__":
    main()


"""
Colorizers

A set of functions to apply colorization to a graphable dataset.
"""

from MLVisualizationTools.backend import GraphData
from MLVisualizationTools.types import ColorizerModes
from typing import Optional
import warnings

def _valid_colorization_call(data: GraphData, colorkey: str, apply_to_model: bool, apply_to_data: bool):
    """Internal function used to throw errors for invalid colorization calls"""
    locations = []
    if apply_to_model:
        locations.append(data.modeldata)
        if data.modeldata.colorized != ColorizerModes.NotColorized:
            warnings.warn("Colorization may overwrite existing colorization.")
    if apply_to_data:
        if data.datavalues is not None:
            locations.append(data.datavalues)
            if data.datavalues.colorized != ColorizerModes.NotColorized:
                warnings.warn("Colorization may overwrite existing colorization.")
        else:
            warnings.warn("Colorization should be applied to data values, but no data was given.")

    if len(locations) == 0:
        warnings.warn("Colorization not being applied to any points")

    if colorkey is not None:
        data.colorkey = colorkey
    data.check_color_key()

    return locations

def simple(data: GraphData, color: str, colorkey: Optional[str] = None,
           apply_to_model:bool = True, apply_to_data: bool = False) -> GraphData:
    """
    Marks all points as being the color inputted

    :param data: GraphData object from a model interface call
    :param color: Color that is applied
    :param colorkey: Overrides default 'Color' as a location to store color
    :param apply_to_model: Applies colorization to model predictions
    :param apply_to_data: Applies colorization to data values
    """

    locations = _valid_colorization_call(data, colorkey, apply_to_model, apply_to_data)

    for d in locations:
        df = d.dataframe

        df[data.colorkey] = color
        d.colorized = ColorizerModes.Simple
        d.basecolor.color = color

    return data

def binary(data: GraphData, highcontrast:bool=True, truecolor: Optional[str] = None, falsecolor: Optional[str] = None,
           cutoff:float=0.5, colorkey='Color', apply_to_model:bool = True, apply_to_data:bool = False) -> GraphData:
    """
    Colors grid based on whether the value is higher than the cutoff. Default colors are green for true and red
    for false. Black will appear if an error occurs.

    :param data: GraphData object from a model interface call
    :param highcontrast: Switches default colors to blue for true and orange for false
    :param truecolor: Manually specify truecolor
    :param falsecolor: Manually specify falsecolor
    :param cutoff: Cutoff value, higher is true
    :param colorkey: Overrides default 'Color' as a location to store color
    :param apply_to_model: Applies colorization to model predictions
    :param apply_to_data: Applies colorization to data values
    """

    if truecolor is None:
        if not highcontrast:
            truecolor = "green"
        else:
            truecolor = "blue"
    if falsecolor is None:
        if not highcontrast:
            falsecolor = "red"
        else:
            falsecolor = "orange"

    locations = _valid_colorization_call(data, colorkey, apply_to_model, apply_to_data)

    for d in locations:
        df = d.dataframe

        df.loc[df[data.outputkey] > cutoff, data.colorkey] = truecolor
        df.loc[df[data.outputkey] <= cutoff, data.colorkey] = falsecolor
        d.colorized = ColorizerModes.Binary
        d.truecolor.color = truecolor
        d.falsecolor.color = falsecolor

    return data


from MLVisualizationTools.backend import fileloader

def getTheme(theme, folder=None, figtemplate=None):
    """
    Backend function for loading theme css files.

    Theme can be 'light' or 'dark', and that will autoload the theme from dbc
    If folder is none, it is set based on the theme
    If figtemplate is none, it is set based on the theme

    Returns theme, folder

    :param theme: 'light' / 'dark' or a css url
    :param folder: path to assets folder
    :param figtemplate: Used for putting plotly in dark theme
    """
    import dash_bootstrap_components as dbc
    if theme == "light":
        theme = dbc.themes.FLATLY
        if folder is None:
            folder = fileloader('theme_assets/light_assets')
        if figtemplate is None:
            figtemplate = "plotly"

    elif theme == "dark":
        theme = dbc.themes.DARKLY
        if folder is None:
            folder = fileloader('theme_assets/dark_assets')
        if figtemplate is None:
            figtemplate = "plotly_dark"

    return theme, folder, figtemplate

def getDashApp(title:str, notebook:bool, usetunneling:bool, host:str, port:int, mode: str, theme, folder):
    """
    Creates a dash or jupyter dash app, returns the app and a function to run it

    :param title: Passed to dash app
    :param notebook: Uses jupyter dash with default port of 1005 (instead of 8050)
    :param usetunneling: Enables ngrok tunneling for kaggle notebooks
    :param host: Passed to dash app.run()
    :param port: Can be used to override default ports
    :param mode: Could be 'inline', 'external', or 'jupyterlab'
    :param theme: Passed to dash app external stylesheets
    :param folder: Passed to assets folder to load theme css
    """

    if port is None:
        if notebook:
            port = 1005
        else:
            port = 8050

    if notebook:
        if usetunneling:
            try:
                from pyngrok import ngrok
            except ImportError:
                raise ImportError("Pyngrok is required to run in a kaggle notebook. "
                                  "Use pip install MLVisualizationTools[kaggle-notebook]")
            ngrok.kill() #disconnects active tunnels
            tunnel = ngrok.connect(port)
            print("Running in an ngrok tunnel. This limits you to 40 requests per minute and one active app.",
                  "For full features use google colab instead.")
            url = tunnel.public_url
        else:
            url = None

        try:
            from jupyter_dash import JupyterDash
        except ImportError:
            raise ImportError("JupyterDash is required to run in a notebook. "
                              "Use pip install MLVisualizationTools[dash-notebook]")
        app = JupyterDash(__name__, title=title, server_url=url,
                               external_stylesheets=[theme], assets_folder=folder)

    else:
        from dash import Dash
        app = Dash(__name__, title=title, external_stylesheets=[theme], assets_folder=folder)

    def runApp():
        if notebook:
            app.run_server(host=host, port=port, mode=mode, debug=True)
        else:
            app.run_server(host=host, port=port, debug=True)

    return app, runApp


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers
from MLVisualizationTools.backend import fileloader
import pandas as pd
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # stops agressive error message printing
from tensorflow import keras

try:
    import matplotlib.pyplot
except ImportError:
    raise ImportError("Matplotlib is required to run this demo. If you don't have matplotlib installed, install it"
                      " with `pip install matplotlib` or run the plotly demo instead.")

def main():
    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))

    AR = Analytics.analyzeModel(model, df, ["Survived"])
    maxvar = AR.maxVariance()
    grid = Interfaces.predictionGrid(model, maxvar[0], maxvar[1], df, ["Survived"])
    grid = Colorizers.simple(grid, 'red')

    plt, _, _ = Graphs.matplotlibGraph(grid, title="Max Variance")
    plt.show(block=False)

    grid = Interfaces.predictionGrid(model, 'Parch', 'SibSp', df, ["Survived"])
    grid = Colorizers.binary(grid, highcontrast=True)
    plt, _, _ = Graphs.matplotlibGraph(grid, title="Parch by SibSp")
    plt.show()

print("This demo shows basic features with tensorflow and matplotlib.")
print("To run the demo, call MatplotlibDemo.main()")

if __name__ == "__main__":
    main()


"""
Model Analytics

A set of functions to anaylze how a ML model responds to different inputs.
All functions return an AnalyticsResult object that contains data about different columns.

Analytics functions can be called directly with a pandas dataframe, or indirectly through
the raw alternative, which takes a colinfo object.
"""


from MLVisualizationTools.backend import colinfo
from typing import List, Dict
import copy
import pandas as pd

#Functions for retrieving data about ml model structure

#TODO - nonlinear

class AnalyticsColumnInfo:
    """Wrapper class for holding col info"""
    def __init__(self, name: str, variance: float):
        self.name = name
        self.variance = variance

    def __lt__(self, other):
        return self.variance < other.variance

    def __repr__(self):
        return "Col with name: " + self.name + " and variance " + str(self.variance)

class AnalyticsResult:
    """Wrapper class for holding and processing col info"""
    def __init__(self):
        self.cols: List[AnalyticsColumnInfo] = []

    def append(self, name: str, variance: float):
        self.cols.append(AnalyticsColumnInfo(name, variance))

    def maxVariance(self):
        """Return a list of cols, ordered by maximum variance"""
        cols = copy.copy(self.cols)
        cols.sort(reverse=True)
        return cols

def analyzeModel(model, data: pd.DataFrame, exclude: List[str] = None, steps:int=20) -> AnalyticsResult:
    """
    Performs 1d analysis on an ML model by calling predict(). Wrapper function for analyzeModelRaw()
    that automatically handles column info generation.

    :param model: A ML model
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    """
    coldata, allcols = colinfo(data, exclude)
    return analyzeModelRaw(model, coldata, allcols, steps)

def analyzeModelRaw(model, coldata: Dict[str, Dict], allcols: List[str], steps:int=20) -> AnalyticsResult:
    """
    Performs 1d analysis on a ML model by calling predict(). Returns a class with lots of info for graphing.
    Call from anaylyzeModel to autogen params.

    Coldata should be formatted with keys 'name', 'min', 'max', 'mean'

    :param model: A ML model
    :param coldata: An dict of dicts accessed by col names with min, max, and mean values
    :param allcols: Ordered list of column names
    :param steps: Resolution to scan model with
    """
    AR = AnalyticsResult()

    predrow = []
    for name, item in coldata.items():
        predrow.append(item['mean'])
    predrow = [predrow] * (steps * len(coldata))
    preddata = pd.DataFrame(predrow, columns=allcols)

    currentpos = 0
    for name, item in coldata.items():
        for i in range(0, steps):
            preddata[name][i + currentpos] = i * (item['max'] - item['min'])/(steps-1) + item['min']
        currentpos += steps

    predictions = model.predict(preddata)

    currentpos = 0
    for name in coldata:
        values = predictions[currentpos:currentpos + steps]
        currentpos += steps
        AR.append(name, values.max() - values.min())
    return AR


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers
from MLVisualizationTools.backend import fileloader
import pandas as pd
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' #stops agressive error message printing
from tensorflow import keras

try:
    import plotly
except ImportError:
    raise ImportError("Plotly is required to run this demo. If you don't have plotly installed, install it with"
                      " `pip install plotly' or run the matplotlib demo instead.")

def main(show=True):
    """Run the demo. Disable show for testing purposes."""
    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))

    AR = Analytics.analyzeModel(model, df, ["Survived"])
    maxvar = AR.maxVariance()

    grid = Interfaces.predictionGrid(model, maxvar[0], maxvar[1], df, ["Survived"])
    grid = Colorizers.binary(grid)
    fig = Graphs.plotlyGraph(grid)
    if show: # pragma: no cover
        fig.show()

    grid = Interfaces.predictionGrid(model, 'Parch', 'SibSp', df, ["Survived"])
    grid = Colorizers.binary(grid, highcontrast=False)
    fig = Graphs.plotlyGraph(grid)
    if show: # pragma: no cover
        fig.show()

print("This demo shows basic features with tensorflow and plotly.")
print("To run the demo, call Demo.main()")

if __name__ == "__main__":
    main()


"""
A set of functions and demos to make machine learning projects easier to understand through effective visualizations.
"""

import MLVisualizationTools.modelanalytics as Analytics
import MLVisualizationTools.modelinterface as Interfaces
import MLVisualizationTools.colorizers as Colorizers
import MLVisualizationTools.graphinterface as Graphs
import MLVisualizationTools.datainterface as DataInterfaces
import MLVisualizationTools.types as types


"""
Data Interface

A set of functions to add datapoints onto an ML response dataset.
"""

from MLVisualizationTools.backend import GraphData, colinfo
from typing import Optional
import pandas as pd

def getHashablePoint(point, graphData, coldata, outputkey, sizekey):
    """Get dictionary information for point clumping"""

    def roundScale(pval, colname, steps):
        minval = coldata[colname]['min']
        maxval = coldata[colname]['max']

        v = round((pval - minval) / ((maxval - minval) / steps)) #scale down to step pos
        return v, round(v * ((maxval - minval) / steps) + minval,6) #scale back to rounded true pos

    sep = ':'

    x, truex = roundScale(point[graphData.x], graphData.x, graphData.steps - 1)
    y, truey = roundScale(point[graphData.y], graphData.y, graphData.steps - 1)
    z, truez = roundScale(point[outputkey], outputkey, graphData.steps - 1)
    pdict = {graphData.x: truex, graphData.y: truey, graphData.outputkey: truez, sizekey: 1}

    hashstr = str(z) + sep + str(x) + sep + str(y)
    if graphData.anim is not None:
        anim, trueanim = roundScale(point[graphData.anim], graphData.anim, graphData.steps - 1)
        hashstr += sep + str(anim)
        pdict[graphData.anim] = trueanim
    return hashstr, pdict

def addClumpedData(graphData: GraphData, dataframe: pd.DataFrame, outputkey: str = 'Output',
                   sizekey: Optional[str] = None) -> GraphData:
    """
    Adds datapoints from dataframe to graphData. Point size is based on frequency of points at that location

    :param graphData: From an interface call
    :param dataframe: Dataframe of points to be added
    :param outputkey: Key in dataframe that represents output of model
    :param sizekey: Used for storing amount of datapoints in a group, must not be in df
    """
    if sizekey is not None:
        graphData.sizekey = sizekey
    graphData.check_size_key()

    if outputkey not in dataframe.columns:
        raise UserWarning(f"Output key '{outputkey}' was not in dataframe. You may need to override the default in the "
                          "function call.")

    coldata, _ = colinfo(dataframe)
    clumpedData = {}

    for _, point in dataframe.iterrows():
        hashstr, pdict = getHashablePoint(point, graphData, coldata, outputkey, graphData.sizekey)
        if hashstr in clumpedData:
            clumpedData[hashstr][graphData.sizekey] += 1
        else:
            clumpedData[hashstr] = pdict

    graphData.add_datavalues(pd.DataFrame.from_dict(clumpedData, orient='index'))
    return graphData

def getHashablePercentagePoint(point, graphData, coldata, outputkey):
    """Get dictionary information for point percentage clumping"""

    def roundScale(pval, colname, steps):
        minval = coldata[colname]['min']
        maxval = coldata[colname]['max']

        v = round((pval - minval) / ((maxval - minval) / steps)) #scale down to step pos
        return v, round(v * ((maxval - minval) / steps) + minval,6) #scale back to rounded true pos

    sep = ':'

    x, truex = roundScale(point[graphData.x], graphData.x, graphData.steps - 1)
    y, truey = roundScale(point[graphData.y], graphData.y, graphData.steps - 1)
    truez = point[outputkey]
    pdict = {graphData.x: truex, graphData.y: truey, graphData.outputkey: [truez]}

    hashstr = str(x) + sep + str(y)
    if graphData.anim is not None:
        anim, trueanim = roundScale(point[graphData.anim], graphData.anim, graphData.steps - 1)
        hashstr += sep + str(anim)
        pdict[graphData.anim] = trueanim
    return hashstr, pdict

def addPercentageData(graphData: GraphData, dataframe: pd.DataFrame, outputkey: str = 'Output',
                      sizekey: Optional[str] = None) -> GraphData:
    """
    Adds datapoints from dataframe to graphData. Point size is based on frequency of points at that location
    Point height is based on the average of point heights at that location

    :param graphData: From an interface call
    :param dataframe: Dataframe of points to be added
    :param outputkey: Key in dataframe that represents output of model
    :param sizekey: Used for storing amount of datapoints in a group, must not be in df
    """
    if sizekey is not None:
        graphData.sizekey = sizekey
    graphData.check_size_key()

    if outputkey not in dataframe.columns:
        raise UserWarning(f"Output key '{outputkey}' was not in dataframe. You may need to override the default in the "
                          "function call.")

    coldata, _ = colinfo(dataframe)
    clumpedData = {}

    for _, point in dataframe.iterrows():
        hashstr, pdict = getHashablePercentagePoint(point, graphData, coldata, outputkey)
        if hashstr in clumpedData:
            clumpedData[hashstr][graphData.outputkey].append(pdict[graphData.outputkey][0])
        else:
            clumpedData[hashstr] = pdict

    for key in clumpedData:
        l = len(clumpedData[key][graphData.outputkey])
        clumpedData[key][graphData.sizekey] = l
        clumpedData[key][graphData.outputkey] = sum(clumpedData[key][graphData.outputkey]) / l

    graphData.add_datavalues(pd.DataFrame.from_dict(clumpedData, orient='index'))
    return graphData




class GraphDataTypes:
    Grid = 'Grid'
    Animation = 'Animation'
    ModelPrediction = 'ModelPrediction'
    DataValues = 'DataValues'

class GraphOutputTypes:
    Auto = 'Auto'
    Matplotlib = 'Matplotlib'
    Plotly = 'Plotly'

class ColorizerModes:
    NotColorized = 'NotColorized'
    Simple = 'Simple'
    Binary = 'Binary'


from werkzeug.serving import is_running_from_reloader

def main(theme='dark', highcontrast=True, mode='external'):
    """
    Runs the demo by calling DashModelVisualizer

    :param theme: theme, could be 'light' or 'dark'
    :param highcontrast: Use blue and orange coloring instead of red and green
    :param mode: where to put the website, could be 'inline' / 'external' / 'jupyterlab'
    """
    from MLVisualizationTools.express import DashModelVisualizer
    from MLVisualizationTools.backend import fileloader
    import pandas as pd
    import os
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # stops agressive error message printing
    from tensorflow import keras

    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))
    df = df.drop("Survived", axis=1)
    DashModelVisualizer.visualize(model, df, title="DashKaggleDemo", theme=theme,
                                  highcontrast=highcontrast, notebook=True, mode=mode, usetunneling=True)

if not is_running_from_reloader():
    print("This demo is for use inside a jupyter notebook that does not supports dash natively (such as kaggle).")
    print("It uses the default precompiled model. To run the demo, call DashKaggleDemo.main()")

if __name__ == "__main__":
    main()


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers, DataInterfaces
from MLVisualizationTools.backend import fileloader
import pandas as pd
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # stops agressive error message printing
from tensorflow import keras
import warnings

hasplotly = False
hasmatplotlib = False

try:
    import plotly
    hasplotly = True
except ImportError:
    try:
        import matplotlib.pyplot
        hasmatplotlib = True
        warnings.warn("Plotly is recommended to run this demo. If you don't have plotly installed, install it"
                      " with `pip install plotly`. Running with matplotlib and limited functionality.")
    except ImportError:
        raise ImportError("Plotly is required to run this demo. If you don't have plotly installed, install it"
                          " with `pip install plotly` or install matplotlib for limited features instead.")
    raise ImportError()

def main(show=True):
    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))

    AR = Analytics.analyzeModel(model, df, ["Survived"])
    maxvar = AR.maxVariance()
    grid = Interfaces.predictionGrid(model, maxvar[0], maxvar[1], df, ["Survived"])
    grid = Colorizers.binary(grid)
    grid = DataInterfaces.addClumpedData(grid, df, 'Survived')

    if hasplotly:
        fig = Graphs.plotlyGraph(grid, title="Clumped Data")
        if show:  # pragma no cover
            fig.show()

    elif hasmatplotlib:
        plt, _, _ = Graphs.matplotlibGraph(grid, title="Clumped Data")
        plt.show(block=False)

    else:
        raise Exception("Plotly and matplotlib missing.")

    grid = Interfaces.predictionGrid(model, maxvar[0], maxvar[1], df, ["Survived"])
    grid = Colorizers.binary(grid)
    grid = DataInterfaces.addPercentageData(grid, df, 'Survived')
    grid = Colorizers.binary(grid, highcontrast=False, apply_to_data=True, apply_to_model=False)

    if hasplotly:
        fig = Graphs.plotlyGraph(grid, title="Percentage Data")
        if show:  # pragma no cover
            fig.show()

    elif hasmatplotlib:
        plt, _, _ = Graphs.matplotlibGraph(grid, title="Percentage Data")
        plt.show(block=False)

    else:
        raise Exception("Plotly and matplotlib missing.")

print("This demo shows data overlay features with plotly.")
print("To run the demo, call DataOverlayDemo.main()")

if __name__ == "__main__":
    main()


import copy

from MLVisualizationTools.types import GraphDataTypes, ColorizerModes
from typing import List, Dict, Tuple, Optional
import warnings
import pandas as pd
from os import path

#Backend functions and classes used by the other scripts

def colinfo(data: pd.DataFrame, exclude:List[str] = None) -> Tuple[Dict[str, Dict], List[str]]:
    """
    Helper function for generating column info dict for a datframe

    :param data: A pandas Dataframe
    :param exclude: A list of data items to exclude
    """
    if exclude is None:
        exclude = []

    coldata = {}
    allcols = []
    for item in data.columns:
        if item not in exclude:
            coldata[item] = {'mean': data[item].mean(), 'min': data[item].min(), 'max': data[item].max()}
            allcols.append(item)
    return coldata, allcols

def fileloader(target: str, dynamic_model_version = True):
    """Specify a path relative to MLVisualizationTools"""
    if dynamic_model_version:
        if 'examples/Models' in target:
            import tensorflow as tf
            if float(tf.version.VERSION[:3]) < 2.5:
                target += "_v2.0"
    return path.dirname(__file__) + '/' + target

class ColorMessage:
    def __init__(self, color: Optional[str], message: str):
        self.color: Optional[str] = color
        self.message: str = message

class ColorizerableDataFrame:
    def __init__(self, dataframe: pd.DataFrame, mode: GraphDataTypes):
        self.dataframe: pd.DataFrame = dataframe

        self.colorized: ColorizerModes = ColorizerModes.NotColorized

        if mode == GraphDataTypes.ModelPrediction:
            self.basecolor = ColorMessage(None, 'Avg. Predictions from Model')
            self.truecolor = ColorMessage(None, 'Avg. Prediction is True')
            self.falsecolor = ColorMessage(None, 'Avg. Prediction is False')

        elif mode == GraphDataTypes.DataValues:
            self.basecolor = ColorMessage(None, 'Actual Data Values')
            self.truecolor = ColorMessage(None, 'Data Values are True')
            self.falsecolor = ColorMessage(None, 'Data Values are False')

        else:
            raise ValueError(str(mode) + " is not a valid data storage mode.")


class GraphData:
    def __init__(self, dataframe: pd.DataFrame, datatype: GraphDataTypes, steps: int, x: str,
                 y: str, anim: str = None, outputkey: str = 'Output'):
        """Class for holding information about grid or animation data to be graphed."""
        self.modeldata: ColorizerableDataFrame = ColorizerableDataFrame(dataframe, GraphDataTypes.ModelPrediction)
        self.datavalues: Optional[ColorizerableDataFrame] = None
        self.datatype: GraphDataTypes = datatype

        self.outputkey: str = outputkey
        self._colorkey: str = 'Color'
        self._sizekey: str = 'Size'

        self.steps: int = steps
        self.x: str = x
        self.y: str = y
        self.anim: Optional[str] = anim
        self.outputkey: str = outputkey

        self.orig_df_cols = copy.deepcopy(list(self.modeldata.dataframe.columns))

    def add_datavalues(self, dataframe: pd.DataFrame):
        self.datavalues = ColorizerableDataFrame(dataframe, GraphDataTypes.DataValues)
        self.orig_df_cols.append(copy.deepcopy(list(self.datavalues.dataframe.columns)))

    @property
    def colorkey(self):
        return self._colorkey

    @colorkey.setter
    def colorkey(self, value: str):
        self._colorkey = value
        self.check_color_key()

    @property
    def sizekey(self):
        return self._sizekey

    @sizekey.setter
    def sizekey(self, value: str):
        self._sizekey = value
        self.check_size_key()

    def check_color_key(self):
        if self._colorkey in self.orig_df_cols:
            warnings.warn(f"Color key '{self._colorkey}' was already in dataframe. This could mean that "
                          f"'{self._colorkey}' was a key in your dataset. This could result in data being overwritten. "
                          "You can pick a different key in the function call.")

    def check_size_key(self):
        if self._sizekey in self.orig_df_cols:
            warnings.warn(f"Size key '{self._sizekey}' was already in dataframe. This means that '{self._sizekey}' was "
                          "a key in your dataset. This could result in data being overwritten. You can pick a different"
                          " key in the function call.")


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers
from MLVisualizationTools.dashbackend import getTheme, getDashApp
import pandas as pd

try:
    import dash
    from dash import Input, Output
    from dash import dcc
    from dash import html
    import dash_bootstrap_components as dbc
    from dash_tour_component import DashTour
    import plotly
except ImportError:
    raise ImportError("Dash and plotly are required to use this tool. Install them with the [dash] flag"
                      " on installation of this library.")

class App:
    def __init__(self, model, data: pd.DataFrame, title:str = "DashModelVisualizer", theme:str = "dark", folder = None,
                 highcontrast:bool = True, notebook:bool = False, usetunneling:bool = False, mode:str = 'external',
                 host:str = '0.0.0.0', port: bool = None):

        theme, folder, self.figtemplate = getTheme(theme, folder)
        self.app, self.runFunc = getDashApp(title, notebook, usetunneling, host, port, mode, theme, folder)

        self.model = model
        self.df = data
        self.highcontrast = highcontrast

        options = []
        for col in self.df.columns:
            options.append({'label': col, 'value': col})

        self.AR = Analytics.analyzeModel(self.model, self.df)
        self.maxvar = self.AR.maxVariance()

        self.x = self.maxvar[0].name
        self.y = self.maxvar[1].name

        self.fig = self.updateGraph()

        graph = dbc.Card([
            dcc.Graph(id='example-graph', figure=self.fig)
        ], body=True)

        config = dbc.Card([
            dbc.Label("X Axis: "),
            dcc.Dropdown(id='xaxis', options=options, value=self.x),
            html.Br(),
            dbc.Label("Y Axis: "),
            dcc.Dropdown(id='yaxis', options=options, value=self.y),
            html.Br(),
            dbc.Button("Open Tour", id='open_tour_button'),
            html.Br(),
        ], body=True)

        tour = DashTour(steps=[{'selector': '[id="xaxis"]',
                                'content': "This dropdown controls the xaxis, it has been preset with the "
                                           "highest variance value."},
                               {'selector': '[id="yaxis"]',
                                'content': "This dropdown controls the yaxis, it has been preset with the "
                                           "second highest variance value."},
                               {'selector': '[id="example-graph"]',
                                'content': "This graph updates when you change the dropdowns."}],
                        isOpen=False,
                        id="tour_component",
                        children=html.Div(),
                        rounded=7
                        )

        self.app.layout = dbc.Container([
            tour,
            html.H1(title),
            html.Hr(),
            dbc.Row([
                dbc.Col(config, md=4),
                dbc.Col(graph, md=8)]
            ),
            html.P()],
            fluid=True,
            className='dash-bootstrap'
        )

        inputs = [Input('xaxis', "value"), Input('yaxis', 'value')]
        self.app.callback(Output("example-graph", "figure"), inputs)(self.updateGraphFromWebsite)
        self.app.callback(Output('tour_component', 'isOpen'), [Input('open_tour_button', 'n_clicks')],
                          prevent_initial_call=True)(lambda _ : True)

    def run(self):
        self.runFunc()

    def updateGraph(self):
        try:
            data = Interfaces.predictionGrid(self.model, self.x, self.y, self.df)
            data = Colorizers.binary(data, highcontrast=self.highcontrast)
            self.fig = Graphs.plotlyGraph(data)
            self.fig.update_layout(template=self.figtemplate)
        except AssertionError:
            pass #User cleared their selection
        return self.fig

    def updateGraphFromWebsite(self, x, y):
        self.x = x
        self.y = y
        return self.updateGraph()

def visualize(model, data: pd.DataFrame, title:str = "DashModelVisualizer", theme:str = "dark", folder = None,
              highcontrast:bool = True, notebook:bool = False, usetunneling:bool = False, mode:str = 'external',
              host:str = '0.0.0.0', port: bool = None):
    """
    Creates a dash website to visualize an ML model.

    :param model: A tensorflow keras model
    :param data: A pandas dataframe, all df columns must be numerical model inputs
    :param title: Title for website
    :param theme: Theme to load app in, can be a string (light / dark) or a url to load a stylesheet from
    :param folder: Directory to load additional css and js from
    :param highcontrast: Visualizes the model with orange and blue instead of green and red. Great for colorblind people!
    :param notebook: Uses jupyter dash instead of dash
    :param usetunneling: Enables ngrok tunneling for use in notebooks that don't have built in dash support, like kaggle
    :param mode: Use 'external', 'inline', or 'jupyterlab'
    :param host: default hostname for dash
    :param port: None for default port (8050) or (1005)
    """
    App(model, data, title, theme, folder, highcontrast, notebook, usetunneling, mode, host, port).run()
